# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""HTML utilities."""

import html

import bleach
from ftfy import fix_text

#: Unwanted unicode characters
UNWANTED_CHARS = {
    # Zero-width space
    u'\u200b',
}


#: Allowed tags used for html sanitizing by bleach.
ALLOWED_HTML_TAGS = [
    'a',
    'abbr',
    'acronym',
    'b',
    'blockquote',
    'br',
    'code',
    'div',
    'em',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'i',
    'li',
    'ol',
    'p',
    'pre',
    'span',
    'strike',
    'strong',
    'sub',
    'sup',
    'u',
    'ul',
]


#: Allowed attributes used for html sanitizing by bleach.
ALLOWED_HTML_ATTRS = {
    '*': ['class'],
    'a': ['href', 'title', 'name', 'class', 'rel'],
    'abbr': ['title'],
    'acronym': ['title'],
}


def strip_html(value):
    """Strip all HTML from text and remove unwanted unicode characters."""
    # Disallow all HTML tags and attributes
    value = sanitize_html(value, tags=[], attrs=[])
    # If value has already escaped HTML then return the unescaped value
    return html.unescape(value)


def sanitize_html(value, tags=None, attrs=None):
    """Sanitizes HTML using the bleach library.

    The default list of allowed tags and attributes is defined by
    ``ALLOWED_HTML_TAGS`` and ``ALLOWED_HTML_ATTRS``.

    You can override the defaults like this:

    .. code-block:: python

        class MySchema(Schema):
            html = fields.SanitizedHTML(tags=['a'], attrs={'a': ['href']})

    :param tags: List of allowed tags.
    :param attrs: Dictionary of allowed attributes per tag.
    """
    value = sanitize_unicode(value)

    if tags is None:
        tags = ALLOWED_HTML_TAGS

    if attrs is None:
        attrs = ALLOWED_HTML_ATTRS

    return bleach.clean(
        value,
        tags=tags,
        attributes=attrs,
        strip=True,
    ).strip()


def is_valid_xml_char(char):
    """Check if a character is valid based on the XML specification."""
    codepoint = ord(char)
    return (0x20 <= codepoint <= 0xD7FF or
            codepoint in (0x9, 0xA, 0xD) or
            0xE000 <= codepoint <= 0xFFFD or
            0x10000 <= codepoint <= 0x10FFFF)


def sanitize_unicode(value, unwanted_chars=None):
    """Sanitize and fix problematic unicode characters."""
    value = fix_text(value.strip())
    # NOTE: This `join` might be ineffiecient... There's a solution with a
    # large compiled regex lying around, but needs a lot of tweaking.
    value = ''.join(filter(is_valid_xml_char, value))

    if unwanted_chars is None:
        unwanted_chars = UNWANTED_CHARS

    for char in unwanted_chars:
        value = value.replace(char, '')
    return value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Marshmallow schemas."""

from .geojson import GeometryObjectSchema, MultiPointSchema, PointSchema, \
    PolygonSchema
from .identifier import IdentifierSchema

__all__ = (
    'GeometryObjectSchema',
    'IdentifierSchema',
    'MultiPointSchema',
    'PointSchema',
    'PolygonSchema',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Link store and field for generating links."""

from collections import defaultdict

from marshmallow import Schema, fields, missing, post_dump
from uritemplate import URITemplate
from werkzeug.datastructures import MultiDict


def _unpack_dict(data):
    """Unpack lists inside a dict.

    For instance this dictionary:

    .. code-block:: python

        {
            "type": ["A", "B"],
            "sort": "newest",
        }

    Is expanded to:

    .. code-block:: python

        [
            ("sort", "newest"),
            ("type", "A"),
            ("type", "B"),
        ]
    """
    for key, value in sorted(data.items()):
        if isinstance(value, list):
            for v in value:
                yield key, v
        else:
            yield key, value


class LinksFactory:
    """Utility class for keeping track of and resolve links.

    :params host: A hostname or a callable returning a hostname.
    :params config: A link configuration.
    """

    def __init__(self, host=None, config=None):
        """Constructor."""
        self._host = host
        self._config = config or {}

    def get_schema(self, namespace, context):
        """Get the schema for a given namespace."""
        schema = self._config.get(namespace)
        if schema:
            return schema(context=context or {})
        return schema

    @property
    def host(self):
        """Get the hostname."""
        return self._host() if callable(self._host) else self._host

    def create_link(self, template, vars):
        """Create the link template with the template variables.

        :param template: URITemplate to expand.
        :param vars: The template variables for the expansion.
        """
        return self.base_url(
            host=self.host,
            rendered_path=template.expand(**self.preprocess_vars(vars))
        )

    @staticmethod
    def preprocess_vars(vars):
        """Preprocess template variables before expansion."""
        for k, v in vars.items():
            if isinstance(v, MultiDict):
                vars[k] = list(sorted(v.items(multi=True)))
            elif isinstance(v, dict):
                # Note, we unpack dicts with list values at this level, because
                # there are no meaningful templates that can make use of them
                # (basically the lists just becomes URL encoded python objects)
                vars[k] = list(_unpack_dict(v))
        return vars

    @staticmethod
    def base_url(scheme="https", host=None, rendered_path="/"):
        """Creates the URL for API and UI endpoints."""
        if host and scheme:
            assert rendered_path.startswith("/")
            return f"{scheme}://{host}{rendered_path}"
        else:
            return rendered_path



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test the marshmallow fields."""

from datetime import date, datetime

import pytest
from marshmallow import EXCLUDE, Schema, ValidationError, missing

from marshmallow_utils import fields


def test_trimmed():
    """Test trimmed string field."""
    class ASchema(Schema):
        f = fields.TrimmedString()

    assert ASchema().load({'f': '   '}) == {'f': ''}
    assert ASchema().load({'f': '  ad sf '}) == {'f': 'ad sf'}


def test_sanitized_unicode():
    """Test sanitized unicode field."""
    class ASchema(Schema):
        f = fields.SanitizedUnicode()

    assert ASchema().load({'f': ' \u200b\u000b\u001b\u0018 '}) == {'f': ''}


def test_sanitized_html():
    """Test sanitized html field."""
    class ASchema(Schema):
        f = fields.SanitizedHTML()

    assert ASchema().load({'f': 'an <script>evil()</script> example'}) == {
        'f': 'an evil() example'}

    # Ensure empty lists to tags/attrs removes all tags.
    class ASchema(Schema):
        f = fields.SanitizedHTML(tags=[], attrs=[])

    assert ASchema().load({'f': '<script>evil()</script><b>Hello</b>'}) == {
        'f': 'evil()Hello'}


def test_stripped_html():
    """Test stripped html field."""
    class ASchema(Schema):
        f = fields.StrippedHTML()

    assert ASchema().dump(
        {'f': 'an <div><span>evil()</span> example</div>'}) == {
        'f': 'an evil() example'}

    # Ensure already escaped HTML is returned unescaped.
    assert ASchema().dump(
        {'f': 'an <div>&lt;span&gt;example&lt;/span&gt;</div>'}) == {
        'f': 'an <span>example</span>'}

    # Unwanted unicode chars from HTML entities should also be removed.
    assert ASchema().dump(
        {'f': 'an &#8203;&quot;example&quot;'}) == {
        'f': 'an "example"'}


def test_isodate():
    """Test ISO date formatted string."""
    class ASchema(Schema):
        f = fields.ISODateString()

    assert ASchema().dump({'f': '1999-10-27'}) == {'f': '1999-10-27'}
    assert ASchema().dump({'f': 'invalid'}) == {}
    assert ASchema().dump({'f': None}) == {}
    assert ASchema().dump({'f': ''}) == {}

    assert ASchema().load({'f': '1999-10-27'}) == {'f': '1999-10-27'}
    pytest.raises(ValidationError, ASchema().load, {'f': 'invalid'})


def test_tzdatetime():
    """Test ISO date formatted string with timezone."""
    class ASchema(Schema):
        f = fields.TZDateTime()

    example_date = datetime(2017, 11, 28, 23, 55, 59, 342380)
    expected_date = '2017-11-28T23:55:59.342380+00:00'
    assert ASchema().dump({'f': example_date}) == {'f': expected_date}

    # Test serialization of None values
    assert ASchema().dump({'f': None}) == {'f': None}


def test_generated():
    """Test fields.generated fields."""

    def serialize_func(obj, ctx):
        return ctx.get('func-foo', obj.get('func-bar', missing))

    def deserialize_func(value, ctx, data):
        return ctx.get('func-foo', data.get('func-bar', missing))

    class GeneratedFieldsSchema(Schema):
        """Test schema."""
        class Meta:
            """Meta attributes for the schema."""

            unknown = EXCLUDE

        gen_function = fields.GenFunction(
            serialize=serialize_func,
            deserialize=deserialize_func,
        )

        gen_method = fields.GenMethod(
            serialize='_serialize_gen_method',
            deserialize='_deserialize_gen_method',
            missing='raises-warning',
        )

        def _serialize_gen_method(self, obj):
            # "meth-foo" from context or "meth-bar" from the object
            return self.context.get(
                'meth-foo', obj.get('meth-bar', missing))

        def _deserialize_gen_method(self, value, data):
            # "meth-foo" from context or "meth-bar" from the data
            return self.context.get(
                'meth-foo', data.get('meth-bar', missing))

    ctx = {
        'func-foo': 'ctx-func-value',
        'meth-foo': 'ctx-meth-value',
    }
    data = {
        'func-bar': 'data-func-value',
        'meth-bar': 'data-meth-value',
        'gen_function': 'original-func-value',
        'gen_method': 'original-meth-value',
    }

    # No context, no data
    assert GeneratedFieldsSchema().load({}) == {}
    assert GeneratedFieldsSchema().dump({}) == {}

    # Only context
    assert GeneratedFieldsSchema(context=ctx).load({}) == {
        'gen_function': 'ctx-func-value',
        'gen_method': 'ctx-meth-value',
    }
    assert GeneratedFieldsSchema(context=ctx).dump({}) == {
        'gen_function': 'ctx-func-value',
        'gen_method': 'ctx-meth-value',
    }

    # Only data
    assert GeneratedFieldsSchema().load(data) == {
        'gen_function': 'data-func-value',
        'gen_method': 'data-meth-value',
    }
    assert GeneratedFieldsSchema().dump(data) == {
        'gen_function': 'data-func-value',
        'gen_method': 'data-meth-value',
    }

    # Context and data
    assert GeneratedFieldsSchema(context=ctx).load(data) == {
        'gen_function': 'ctx-func-value',
        'gen_method': 'ctx-meth-value',
    }
    assert GeneratedFieldsSchema(context=ctx).dump(data) == {
        'gen_function': 'ctx-func-value',
        'gen_method': 'ctx-meth-value',
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test the marshmallow ISO Language."""

import pytest
from marshmallow import Schema, ValidationError

from marshmallow_utils.fields import ISOLangString


class MySchema(Schema):
    f = ISOLangString()


def test_iso639_3_assert():

    assert MySchema().load({'f': "tes"}) == {'f': "tes"}


@pytest.mark.parametrize('test_input', ["te", "12t", "te!", ",te!"])
def test_iso639_3_raise(test_input):

    pytest.raises(ValidationError, MySchema().load, {'f': test_input})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Trimming string field."""

from marshmallow import fields


class TrimmedString(fields.String):
    """String field which strips whitespace at the ends of the string."""

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize string value."""
        value = super()._deserialize(
            value, attr, data, **kwargs)
        return value.strip()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

from datetime import date

from marshmallow import Schema, fields

from marshmallow_utils.fields import nestedattr


class ArtistSchema(Schema):
    """Artist Schema."""
    name = fields.Str()


class AlbumSchema(Schema):
    """Album Schema."""

    artist = nestedattr.NestedAttribute(ArtistSchema())


class MyAlbum(dict):

    @property
    def artist(self):
        return {'name': self['artist']['name'].capitalize()}


def test_schema():
    album = MyAlbum(
        artist={'name': 'david'},  # this shouldn't be used in the schema dump
        title="Hunky Dory",
        release_date=date(1971, 12, 17)
    )

    schema = AlbumSchema()
    result = schema.dump(album)

    assert result['artist'] == {'name': 'David'}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Sanitized Unicode string field."""

from marshmallow import fields

from ..html import sanitize_unicode


class SanitizedUnicode(fields.String):
    """String field that sanitizes and fixes problematic unicode characters."""

    UNWANTED_CHARACTERS = {
        # Zero-width space
        u'\u200b',
    }

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize sanitized string value."""
        value = super()._deserialize(
            value, attr, data, **kwargs)
        return sanitize_unicode(value, unwanted_chars=self.UNWANTED_CHARACTERS)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test the localization of EDTF string."""

import pytest
from marshmallow import Schema, ValidationError

from marshmallow_utils.fields import EDTFDateString


class TestSchema(Schema):
    date = EDTFDateString()


def test_dump():
    assert TestSchema().dump({'date': '2020-09/2020-10'}) == {
        'date': '2020-09/2020-10',
    }


def test_load():
    s = TestSchema()
    assert s.load({'date': '2020-09/2020-10'})
    # Invalid
    pytest.raises(ValidationError, s.load, {'date': '2020-09-21garbage'})
    # Not chronological
    pytest.raises(ValidationError, s.load, {'date': '2021/2020'})
    # Not date or interval
    pytest.raises(ValidationError, s.load, {'date': '2020-01-01T10:00:00'})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Localized Extended Date(/Time) Format Level 0 date string field."""

import calendar
from datetime import date, datetime

import arrow
from babel import Locale
from babel.core import negotiate_locale
from babel.dates import LC_TIME, format_date, format_datetime, format_time
from babel_edtf import format_edtf
from marshmallow import fields


class BabelFormatField(fields.String):
    """Base classe for babel date and time formatting fields.

    The babel format field is used only for dumping.
    """

    def __init__(self, format='medium', locale=LC_TIME, parse=True, **kwargs):
        """Constructor.

        :param format: The format to use (either ``short``, ``medium``,
                       ``long`` or ``full``).
        :param locale: The current locale or a callable returning the current
                       locale.
        """
        self._format = format
        self._locale = locale
        self._parse = parse
        kwargs.setdefault('dump_only', True)
        super().__init__(**kwargs)

    @property
    def locale(self):
        """Get the locale to use."""
        return self._locale() if callable(self._locale) else self._locale

    def parse(self, value, as_time=False, as_date=False, as_datetime=False):
        """Parse the value if it's a string."""
        if not self._parse or not isinstance(value, str):
            return value

        a = arrow.get(value)
        if as_time:
            return a.datetime
        elif as_date:
            return a.date()
        elif as_datetime:
            return a.datetime
        else:
            return a

    def format_value(self, value):
        """Format a given value using the chosen format function."""
        raise NotImplementedError()

    def _serialize(self, value, attr, data, **kwargs):
        """Serialize the value."""
        return super()._serialize(
            self.format_value(value), attr, data, **kwargs)


class FormatDate(BabelFormatField):
    """Format a date object."""

    def format_value(self, value):
        """Format an EDTF date."""
        return format_date(
            self.parse(value, as_date=True),
            format=self._format,
            locale=self.locale
        )


class FormatDatetime(BabelFormatField):
    """Format a datetime object."""

    def __init__(self, tzinfo=None, **kwargs):
        """Constructor."""
        self._tzinfo = tzinfo
        super().__init__(**kwargs)

    @property
    def tzinfo(self):
        """Get the timzone to use."""
        return self._tzinfo() if callable(self._tzinfo) else self._tzinfo

    def format_value(self, value):
        """Format an EDTF date."""
        return format_datetime(
            self.parse(value, as_datetime=True),
            format=self._format,
            tzinfo=self.tzinfo,
            locale=self.locale
        )


class FormatTime(FormatDatetime):
    """Format a time object."""

    def format_value(self, value):
        """Format an EDTF date."""
        return format_time(
            self.parse(value, as_time=True),
            format=self._format,
            tzinfo=self.tzinfo,
            locale=self.locale
        )


class FormatEDTF(BabelFormatField):
    """Format an EDTF-formatted string."""

    def format_value(self, value):
        """Format an EDTF date."""
        return format_edtf(value, format=self._format, locale=self.locale)


class BabelGettextDictField(fields.String):
    """Translation string field (dump only).

    This field dumps a translation string as output, by looking up the
    translation in the dictionary provided as input (the message catalog).

    The lookup is performed via babel's locale negotiation (e.g. en_US will
    also match en).

    Basically the fields takes a data object like this::

        {'title': {'en': 'Text', 'da': 'Tekst'}}

    and dumps this (in case the locale is english)::

        {'title': 'Text'}
    """

    default_error_messages = {
        'invalid': 'Not a valid dictionary.',
        'missing_locale': 'Translation not found for ',
    }

    def __init__(self, locale, default_locale, **kwargs):
        """Initialize the field.

        :param locale: The locale to lookup, or a function returning the
            locale.
        :param default_locale: The default locale in case the locale is not
            found. Can be a callable that returns the default locale.
        """
        self._locale = locale
        self._default_locale = default_locale
        kwargs['dump_only'] = True
        super().__init__(**kwargs)

    @property
    def locale(self):
        """Get the locale to be used."""
        return self._locale() if callable(self._locale) else self._locale

    @property
    def default_locale(self):
        """Get the default locale to be used."""
        return self._default_locale() \
            if callable(self._default_locale) else self._default_locale

    def _serialize(self, value, attr, obj, **kwargs):
        """Serialize the dict into a string.

        The dict is a message catalog with the keys being locale identifiers
        and the values being the translated string.
        """
        if value is None:
            return None
        if not isinstance(value, dict):
            raise self.make_error('invalid')
        translated_str = gettext_from_dict(
            value, self.locale, self.default_locale)
        if translated_str is None:
            raise self.make_error('missing_locale')
        return super()._serialize(translated_str, attr, obj, **kwargs)


def gettext_from_dict(catalog, locale, default_locale):
    """Get translation string from a dictionary."""
    # First try with negotiate_locale. Negotiate locale will not properly
    # negotiate e.g "en" when the available locales are "en_GB" and "da", even
    # though "en_GB" could be used.
    selected = negotiate_locale([str(locale)], catalog.keys())
    if selected:
        return catalog[selected]

    # In situations where negotiate locale doesn't work, we check if the
    # language itself might be found.

    # Extract language keys only.
    catalog_langs = {
        Locale.parse(l).language: l for (l, msg) in catalog.items()
    }
    if isinstance(locale, str):
        locale = Locale.parse(locale)
    if locale.language in catalog_langs:
        # If primary language match, use that
        catalog_key = catalog_langs[locale.language]
        return catalog[catalog_key]
    # If not, use default locale (must be defined it is defined)
    # "en" is set as fallback lng.
    return catalog.get(str(default_locale), "en")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for marshmallow Identifiers field."""

from functools import partial

import idutils
import pytest
from marshmallow import Schema, ValidationError
from marshmallow.fields import Nested

from marshmallow_utils.fields import IdentifierSet
from marshmallow_utils.schemas import IdentifierSchema


class TestSchema(Schema):
    allowed_schemes = {
        "orcid": {"label": "ORCID", "validator": idutils.is_orcid},
        "doi": {"label": "DOI", "validator": idutils.is_doi}
    }
    identifiers = IdentifierSet(
        Nested(
            partial(IdentifierSchema, allowed_schemes=allowed_schemes)
        )
    )


def test_valid_identifiers():
    valid_identifiers = {
        "identifiers": [{
            "identifier": "0000-0001-6759-6273",
            "scheme": "orcid"
            }, {
            "identifier": "10.5281/zenodo.9999999",
            "scheme": "doi"
        }]
    }

    assert TestSchema().load(valid_identifiers) == valid_identifiers


def test_invalid_duplicate_identifiers():
    duplicate_identifiers = [
        {
            "identifiers": [{  # full duplication
                "identifier": "10.5281/zenodo.9999999",
                "scheme": "doi"
                }, {
                "identifier": "10.5281/zenodo.9999999",
                "scheme": "doi"
            }]
        }, {
            "identifiers": [{  # no scheme provided
                "identifier": "10.5281/zenodo.9999999",
                }, {
                "identifier": "10.5281/zenodo.9999999",
            }]
        }
    ]

    for case in duplicate_identifiers:
        with pytest.raises(ValidationError) as e:
            TestSchema().load(case)
        errors = e.value.normalized_messages()
        assert errors == {
            'identifiers': ['Only one identifier per scheme is allowed.']
        }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Contributed Marshmallow fields."""

import functools
from inspect import getfullargspec, isfunction, ismethod

from marshmallow import fields, utils


def _get_func_args(func):
    """Get a list of the arguments a function or method has."""
    if isinstance(func, functools.partial):
        return _get_func_args(func.func)
    if isfunction(func) or ismethod(func):
        return list(getfullargspec(func).args)
    if callable(func):
        return list(getfullargspec(func.__call__).args)


class Function(fields.Function):
    """Enhanced marshmallow Function field.

    The main difference between the original marshmallow.fields.Function is for
    the ``deserialize`` function, which can now also point to a three-argument
    function, with the third argument being the original data that was passed
    to ``Schema.load``. The following example better demonstrates how this
    works:

    .. code-block:: python

        def serialize_foo(obj, context):
            return {'serialize_args': {'obj': obj, 'context': context}}

        def deserialize_foo(value, context, data):
            return {'deserialize_args': {
                'value': value, 'context': context, 'data': data}}

        class FooSchema(marshmallow.Schema):

            foo = Function(serialize_foo, deserialize_foo)

        FooSchema().dump({'foo': 42})
        {'foo': {
            'serialize_args': {
                'obj': {'foo': 42},
                'context': {}  # no context was passed
            }
        }}

        FooSchema().load({'foo': 42}).data
        {'foo': {
            'deserialize_args': {
                'value': 42,
                'context': {},  # no context was passed
                'data': {'foo': 42},
            }
        }}
    """

    def _deserialize(self, value, attr, data, **kwargs):
        if self.deserialize_func:
            return self._call_or_raise(
                self.deserialize_func, value, attr, data)
        return value

    def _call_or_raise(self, func, value, attr, data=None):
        func_args_len = len(_get_func_args(func))
        if func_args_len > 2:
            return func(value, self.parent.context, data)
        elif func_args_len > 1:
            return func(value, self.parent.context)
        else:
            return func(value)


class Method(fields.Method):
    """Enhanced marshmallow Method field.

    The main difference between the original marshmallow.fields.Method is for
    the ``deserialize`` method, which can now also point to a two-argument
    method, with the second argument being the original data that was passed to
    ``Schema.load``. The following example better demonstrates how this works:

    .. code-block:: python

        class BarSchema(marshmallow.Schema):

            bar = Method('serialize_bar', 'deserialize_bar')

            # Exactly the same behavior as in ``marshmallow.fields.Method``
            def serialize_bar(self, obj):
                return {'serialize_args': {'obj': obj}}

            def deserialize_bar(self, value, data):
                return {'deserialize_args': {'value': value, 'data': data}}

        BarSchema().dump({'bar': 42})
        {'bar': {
            'serialize_args': {
                'obj': {'bar': 42}
            }
        }}

        BarSchema().load({'bar': 42})
        {'bar': {
            'deserialize_args': {
                'data': {'bar': 42},
                'value': 42
            }
        }}
    """

    def _deserialize(self, value, attr, data, **kwargs):
        if self.deserialize_method_name:
            try:
                method = utils.callable_or_raise(
                    getattr(self.parent, self.deserialize_method_name, None)
                )
                method_args_len = len(_get_func_args(method))
                if method_args_len > 2:
                    return method(value, data)
                return method(value)
            except AttributeError:
                pass
        return value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test LinksStore."""

import pytest
from marshmallow import Schema
from uritemplate import URITemplate

from marshmallow_utils import fields
from marshmallow_utils.links import LinksFactory


@pytest.fixture()
def my_schema():
    """A test schema with."""
    class LinksSchema(Schema):
        self = fields.Link(
            template=URITemplate("/records{?params*}"),
            params=lambda o: {"params": {
                # type is expected to show in the query string as type=A&type=B
                "type": ["A", "B"],
                "sort": "newest",
                "subtype": ["1"],
                "size": 10,
            }}
        )
        publish = fields.Link(
            template=URITemplate("/admin{/pid}"),
            params=lambda o: {"pid": o.get("pid")},
            permission="admin",
        )
        prev = fields.Link(
            template=URITemplate("/prev"),
            params=lambda o: {},
            when=lambda o: o.get("allowed", True)
        )

    class MySchema(Schema):
        links = fields.Links()

    factory = LinksFactory(
        host="localhost",
        config={"search": LinksSchema}
    )

    return MySchema(
        context={
            "links_factory": factory,
            "links_namespace": "search"

        }
    )


def test_links():
    """Test links factory with links field."""
    class EntitySchema(Schema):
        links = fields.Links()

    class ItemLinkSchema1(Schema):
        self = fields.Link(
            template=URITemplate("/1/{?pid}"),
            params=lambda o: {'pid': o.get("pid")},
            permission="read"
        )

    class ItemLinkSchema2(Schema):
        self = fields.Link(
            template=URITemplate("/2/{?pid}"),
            params=lambda o: {'pid': o.get("pid")},
            permission="create"
        )

    # Test: Create links during dumping
    f = LinksFactory(host="localhost", config={"item": ItemLinkSchema1})
    context = {
        "links_factory": f,
        "links_namespace": "item",
        "field_permission_check": lambda o: True  # allows anything
    }
    assert (
        {'links': {'self': 'https://localhost/1/?pid=1'}} ==
        EntitySchema(context=context).dump({"pid": 1})
    )

    # Test: Create links with other config during dumping
    f = LinksFactory(host="localhost", config={"item": ItemLinkSchema2})
    context = {
        "links_factory": f,
        "links_namespace": "item",
        "field_permission_check": lambda o: True  # allows anything
    }
    assert (
        {'links': {'self': 'https://localhost/2/?pid=1'}} ==
        EntitySchema(context=context).dump({"pid": 1})
    )


def test_params_expansion(my_schema):
    """Test link store resolver."""
    # Test self link generation with expansion of the "type" query string
    # argument.
    self_link = my_schema.dump({})["links"]["self"]
    assert self_link == \
        "https://localhost/records?size=10&sort=newest&subtype=1&type=A&type=B"


def test_unknown_namespace(my_schema):
    """Test unknown namespace."""
    my_schema.context["links_namespace"] = "unknown"
    assert my_schema.dump({})["links"] == {}


def test_permission(my_schema):
    """Test permission checks."""
    my_schema.context["field_permission_check"] = lambda a: a != "admin"
    links = my_schema.dump({})["links"]
    assert 'self' in links
    assert 'publish' not in links

    my_schema.context["field_permission_check"] = lambda a: True
    links = my_schema.dump({})["links"]
    assert 'self' in links
    assert 'publish' in links


def test_when(my_schema):
    links = my_schema.dump({"allowed": False})["links"]
    assert 'self' in links
    assert 'publish' in links
    assert 'prev' not in links

    links = my_schema.dump({"allowed": True})["links"]
    assert 'self' in links
    assert 'publish' in links
    assert 'prev' in links


def test_link_field_initialization():
    """Test that you can pass both URI templates and strings."""
    tpls = ["/records{?params*}", URITemplate("/records{?params*})")]
    for t in tpls:
        assert isinstance(fields.Link(template=t).template, URITemplate)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Persistent identifier Schema with automatic scheme detection.

Integrates idutils library to detect the scheme.
"""

import idutils
from marshmallow import Schema, ValidationError, post_load, pre_load, \
    validates_schema

from ..fields import SanitizedUnicode


class IdentifierSchema(Schema):
    """Identifier with automatic scheme detection."""

    identifier = SanitizedUnicode()
    scheme = SanitizedUnicode()

    error_messages = {
        "invalid_identifier": "Invalid {scheme} identifier.",
        "invalid_scheme": "Invalid scheme.",
        "required": "Missing data for required field.",
    }

    def __init__(self, allowed_schemes, identifier_required=True, **kwargs):
        """Constructor.

        :param allowed_schemes: a dictionary of allowed schemes. Each key must
            contain a validator function and a scheme label.
        :param identifier_required: True when the identifier value is required.
        """
        self.identifier_required = identifier_required
        self.allowed_schemes = allowed_schemes
        super().__init__(**kwargs)

    def _intersect_with_order(self, detected_schemes):
        """Returns the first detected scheme that is allowed."""
        allowed_schemes = set(self.allowed_schemes.keys())
        for detected in detected_schemes:
            if detected in allowed_schemes:
                return detected

        return None

    @pre_load(pass_many=False)
    def load_scheme(self, data, **kwargs):
        """Loads the scheme of the identifier."""
        # If no identifier provided, proceed to validation
        identifier = data.get("identifier")
        if not identifier:
            return data

        # If identifier and scheme is provided, proceed to validation.
        scheme = data.get("scheme")
        if scheme:
            return data

        # If identifier but no scheme is provided, try to detect scheme
        detected_schemes = idutils.detect_identifier_schemes(identifier)

        # Select a valid scheme from the detected ones.
        detected_scheme = self._intersect_with_order(detected_schemes)
        if detected_scheme:
            data["scheme"] = detected_scheme
        return data

    @validates_schema
    def validate_identifier(self, data, **kwargs):
        """Validate the identifier format and scheme."""
        identifier = data.get("identifier")
        scheme = data.get("scheme")

        # Bail if identifier is not required and identifier/scheme is not
        # provided
        if not self.identifier_required and not identifier and not scheme:
            return

        errors = dict()

        # Validate scheme
        if not scheme:
            errors['scheme'] = self.error_messages["required"]
        elif scheme not in self.allowed_schemes:
            errors['scheme'] = self.error_messages["invalid_scheme"]

        # Validate identifier
        if not identifier:
            errors['identifier'] = self.error_messages["required"]
        elif scheme and scheme in self.allowed_schemes:
            validator = self.allowed_schemes[scheme]["validator"]
            if not validator(identifier):
                message = self.error_messages["invalid_identifier"]
                scheme_label = self.allowed_schemes[scheme].get(
                    "label", scheme
                )
                errors['identifier'] = message.format(scheme=scheme_label)

        if errors:
            raise ValidationError(errors)

    @post_load
    def normalize_identifier(self, data, **kwargs):
        """Normalizes the identifier based on the scheme."""
        identifier = data.get("identifier")

        # It can be empty if not required
        if identifier:
            # at this point, `scheme` is set or validation failed earlier
            scheme = data["scheme"]
            # will return the same value if not able to normalize by idutils
            data["identifier"] = idutils.normalize_pid(identifier, scheme)

        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for marshmallow GeoJSON schema."""

import pytest
from marshmallow import ValidationError

from marshmallow_utils.schemas import GeometryObjectSchema


def test_point():
    valid_full = {
        "type": "Point",
        "coordinates": [-32.94682, -60.63932]
    }

    point = GeometryObjectSchema().load(valid_full)
    assert valid_full == point == GeometryObjectSchema().dump(point)


def test_point_fail():
    invalid_input = {
        "type": "Point",
        "coordinates": [-32.94682]
    }
    pytest.raises(ValidationError, GeometryObjectSchema().load, invalid_input)


def test_multipoint():
    valid_full = {
        "type": "MultiPoint",
        "coordinates": [[-32.94682, -60.63932], [-32.94682, -60.63932, 10.0]]
    }

    multipoint = GeometryObjectSchema().load(valid_full)
    assert valid_full == multipoint == GeometryObjectSchema().dump(multipoint)


def test_multipoint_fail():
    invalid_input = {
        "type": "MultiPoint",
        "coordinates": [-32.94682, -60.63932]
    }
    pytest.raises(ValidationError, GeometryObjectSchema().load, invalid_input)


def test_polygon():
    valid_full = {
        "type": "Polygon",
        "coordinates": [[
            [2.38, 57.322], [23.194, -20.28], [-120.43, 19.15], [2.38, 57.322]
        ]]
    }

    polygon = GeometryObjectSchema().load(valid_full)
    assert valid_full == polygon == GeometryObjectSchema().dump(polygon)


def test_polygon_fail():
    invalid_input = {
        "type": "Polygon",
        "coordinates": [[[2.38, 57.322], [23.194, -20.28], [2.38, 57.322]]]
    }
    pytest.raises(ValidationError, GeometryObjectSchema().load, invalid_input)


@pytest.mark.parametrize("geotype", [("Point"), ("MultiPoint")])
def test_point_no_coordinates_fail(geotype):
    invalid_point = {"type": geotype}

    pytest.raises(ValidationError, GeometryObjectSchema().load, invalid_point)


def test_not_supported_type_fail():
    invalid_type = {"type": "invalid"}

    pytest.raises(ValidationError, GeometryObjectSchema().load, invalid_type)


def test_no_type_fail():
    invalid_no_type = {
        "coordinates": [-32.94682, -60.63932]
    }

    pytest.raises(
        ValidationError, GeometryObjectSchema().load, invalid_no_type)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test the marshmallow contrib."""

import pytest
from marshmallow import Schema, ValidationError

from marshmallow_utils.fields import Function, Method


def serialize_foo(obj, context):
    return {'serialize_args': {'obj': 1, 'context': 2}}


def deserialize_foo(value, context, data):
    return {'deserialize_args': {
        'value': 3, 'context': 4, 'data': 5}}


class MySchema(Schema):
    f = Function(serialize_foo, deserialize_foo)


def test_contrib_function_field_dump():
    assert MySchema().dump({'f': 12123}) == {'f': {'serialize_args':
                                             {'context': 2, 'obj': 1}}}


def test_contrib_function_field_load():
    assert MySchema().load({'f': 12123}) == {'f': {'deserialize_args': {
                                            'value': 3, 'context': 4,
                                            'data': 5}}}


class BarSchema(Schema):
    bar = Method('serialize_bar', 'deserialize_bar')

    def serialize_bar(self, obj):
        return {'serialize_args': {'obj': 1}}

    def deserialize_bar(self, value, data):
        return {'deserialize_args': {'value': 23, 'data': 24}}


def test_contrib_method_field_dump():
    assert BarSchema().dump({'bar': 12123}) == {'bar': {'serialize_args':
                                                {'obj': 1}}}


def test_contrib_method_field_load():
    assert BarSchema().load({'bar': 12123}) == {'bar': {'deserialize_args':
                                                {'value': 23, 'data': 24}}}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Extended Date(/Time) Format Level 0 date string field."""

from edtf import Date, Interval, parse_edtf
from edtf.parser.grammar import ParseException
from marshmallow import ValidationError, fields
from marshmallow.validate import Validator


class EDTFValidator(Validator):
    """EDTF validator."""

    default_message = "Please provide a valid date or interval."

    def __init__(self, types=[Date, Interval], chronological_interval=True,
                 error=None):
        """Constructor.

        :params types: List of EDTFObject subclasses that you accept. Use
            EDTFObject to accept all levels.
        """
        self._types = types or []
        self._chronological_interval = chronological_interval
        self._error = error or self.default_message

    def _format_error(self, value, e):
        return self._error.format(input=value, edtf=e)

    def __call__(self, value):
        """Validate."""
        try:
            e = parse_edtf(value)
        except ParseException:
            raise ValidationError(self._format_error(value, None))

        if self._types:
            if not any([isinstance(e, t) for t in self._types]):
                raise ValidationError(self._format_error(value, e))

        if self._chronological_interval:
            # We require intervals to be chronological. EDTF Date and Interval
            # both have same interface and
            # date.lower_strict() <= date.upper_strict() is always True for a
            # Date
            if e.upper_strict() < e.lower_strict():
                raise ValidationError(self._format_error(value, e))

        return value


class EDTFDateString(fields.Str):
    """
    Extended Date(/Time) Format Level 0 date string field.

    A string field which an using the EDTF Validator.
    """

    def __init__(self, **kwargs):
        """Constructor."""
        kwargs.setdefault('validate', EDTFValidator())
        super().__init__(**kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for marshmallow GeoJSON Point schema."""

import pytest
from marshmallow import ValidationError

from marshmallow_utils.schemas import PointSchema


@pytest.mark.parametrize("coordinates", [
    ([-32.94682, -60.63932]),
    ([-32.94682, -60.63932, 10.0]),
])
def test_point(coordinates):
    valid_full = {
        "type": "Point",
        "coordinates": coordinates
    }

    loaded = PointSchema().load(valid_full)
    # NOTE: Since the schemas return the dict itself, the loaded object
    # is the same than the input and dumped objects (dicts)
    assert valid_full == loaded == PointSchema().dump(loaded)


# NOTE: ["-32.94682", "-60.63932"] is valid because the Float field
# deserializes properly the string into float, so it becomes
# [-32.94682, -60.63932]
@pytest.mark.parametrize("coordinates", [
    ("-32.94682,-60.63932"),
    (["-32.94682,-60.63932"]),
    ([-32.94682, -60.63932, 10.0, 10.0]),
    ([-32.94682])
])
def test_point_fail(coordinates):
    invalid_point = {
        "type": "Point",
        "coordinates": coordinates
    }
    pytest.raises(ValidationError, PointSchema().load, invalid_point)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for marshmallow GeoJSON MultiPoint schema."""

import pytest
from marshmallow import ValidationError

from marshmallow_utils.schemas import MultiPointSchema


@pytest.mark.parametrize("coordinates", [
    ([[-32.94682, -60.63932], [-32.94682, -60.63932, 10.0]]),
    ([[-32.99, -60.63], [-32.94, -60.63], [-32.92, -60.32, 10.0]]),
])
def test_multipoint(coordinates):
    valid_full = {
        "type": "MultiPoint",
        "coordinates": coordinates
    }
    loaded = MultiPointSchema().load(valid_full)
    # NOTE: Since the schemas return the dict itself, the loaded object
    # is the same than the input and dumped objects (dicts)
    assert valid_full == loaded == MultiPointSchema().dump(loaded)


@pytest.mark.parametrize("coordinates", [
    ("-32.94682,-60.63932, -32.94682, -60.63932, 10.0"),
    (["-32.94682,-60.63932", "-32.94682, -60.63932, 10.0"]),
    ([-32.94682, -60.63932, 10.0, 10.0]),
    ([[-32.99, -60.63], [-32.94], [-32.92, -60.32, 10.0]])
])
def test_multipoint_fail(coordinates):
    invalid_point = {
        "type": "MultiPoint",
        "coordinates": coordinates
    }
    pytest.raises(ValidationError, MultiPointSchema().load, invalid_point)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Generated field."""

import warnings

from marshmallow import __version_info__ as marshmallow_version
from marshmallow import missing as missing_

from .contrib import Function, Method


class GeneratedValue(object):
    """Sentinel value class forcing marshmallow missing field generation."""

    pass


class ForcedFieldDeserializeMixin(object):
    """Mixin that forces deserialization of marshmallow fields."""

    # Overriding default deserializer since we need to deserialize an
    # initially non-existent field. In this implementation the checks are
    # removed since we expect our deserializer to provide the value.
    def deserialize(self, *args, **kwargs):
        """Deserialize field."""
        # Proceed with _deserialization, skipping all checks.
        output = self._deserialize(*args, **kwargs)
        self._validate(output)
        return output


class GenFunction(ForcedFieldDeserializeMixin, Function):
    """Function field which is always deserialized."""

    pass


class GenMethod(ForcedFieldDeserializeMixin, Method):
    """Method field which is always deserialized."""

    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Datetime field which converts naive datetimes to TZ aware datetimes."""

from datetime import timezone

from marshmallow import fields


class TZDateTime(fields.DateTime):
    """Datetime field which converts naive datetimes to TZ aware datetimes.

    Defaults to setting the timezone to UTC, and using ISO format.
    """

    def __init__(self, timezone=timezone.utc, format='iso', **kwargs):
        """Initialize the field."""
        super().__init__(format=format, **kwargs)
        self.timezone = timezone

    def _serialize(self, value, attr, obj, **kwargs):
        """Serialize a datetime to add the timezone (UTC)."""
        if value is not None:
            value = value.replace(tzinfo=self.timezone)
        return super()._serialize(value, attr, obj, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test the babel date and time localization."""

from datetime import date, datetime

import pytest
from babel import Locale
from babel.dates import get_timezone
from marshmallow import Schema, ValidationError

from marshmallow_utils.fields import BabelGettextDictField, FormatDate, \
    FormatDatetime, FormatEDTF, FormatTime
from marshmallow_utils.fields.babel import gettext_from_dict


@pytest.fixture()
def dt():
    """Datetime fixture."""
    return datetime(2020, 11, 8, 23, 22)


class MySchema(Schema):
    """Test schema."""

    short = FormatEDTF(attribute='edtf', format='short', locale='en')
    long = FormatEDTF(attribute='edtf', format='long', locale=lambda: 'en')
    datetime = FormatDatetime(format='medium', locale='da')
    time = FormatTime(format='long', locale='da',
                      tzinfo=lambda: get_timezone('America/Chicago'))
    date = FormatDate(format='short', locale='da')
    no_parse = FormatDatetime(format='medium', locale='da', parse=False)


def test_format_edtf():
    """Test EDTF formatting."""
    assert MySchema().dump({'edtf': '2020-09/2020-10'}) == {
        'short': '9/2020 – 10/2020',
        'long': 'September – October 2020',
    }


def test_format_datetime(dt):
    """Test datetime formatting."""
    assert MySchema().dump({'datetime': dt}) == {
        'datetime': '8. nov. 2020 23.22.00',
    }
    assert MySchema().dump({'datetime': dt.isoformat()}) == {
        'datetime': '8. nov. 2020 23.22.00',
    }
    # Test field with no parsing of strings:
    pytest.raises(Exception, MySchema().dump, {'no_parse': dt.isoformat()})


def test_format_time(dt):
    """Test datetime formatting."""
    assert MySchema().dump({'time': dt}) == {'time': '17.22.00 -0600'}
    assert MySchema().dump({
        'time': dt.isoformat()}) == {'time': '17.22.00 -0600'}


def test_format_date(dt):
    """Test datetime formatting."""
    assert MySchema().dump({'date': dt}) == {'date': '08.11.2020'}
    assert MySchema().dump({'date': '2021-01-01'}) == {'date': '01.01.2021'}


def test_babelgettextdictfield():
    """Test the babel gettext dict field."""
    class TestSchema(Schema):
        title = BabelGettextDictField(locale='da', default_locale='en')
    s = TestSchema()
    # Success
    assert s.dump({'title': {'en': 'EN', 'da': 'DA'}}) == \
        {'title': 'DA'}
    assert s.dump({'title': {'en': 'EN'}}) == \
        {'title': 'EN'}
    assert s.dump({'title': None}) == {'title': None}

    # Fail - invalid data type
    assert pytest.raises(ValidationError, s.dump, {'title': 'invalid'})

    # Fail - default language not in catalog
    # assert pytest.raises(ValidationError, s.dump, {'title': {'de': 'DE'}})


def test_babelgettextdictfield_functions():
    """Test the babel gettext dict field (with functions and parameters)."""
    class TestSchema(Schema):
        title = BabelGettextDictField(
            locale=lambda: Locale.parse('en'),
            default_locale=lambda: Locale.parse('da'))
    s = TestSchema()

    assert s.dump({'title': {'en': 'EN', 'da': 'DA'}}) == \
        {'title': 'EN'}
    assert s.dump({'title': {'da': 'DA'}}) == \
        {'title': 'DA'}
    assert s.dump({'title': None}) == {'title': None}

    # Fail - default language not in catalog
    # assert pytest.raises(ValidationError, s.dump, {'title': {'de': 'DE'}})


def test_gettext_from_dict():
    """Test the locale negotiation."""
    assert gettext_from_dict({'en': 'en', 'en_US': 'en_US'}, 'en_US', 'da') \
        == 'en_US'
    assert gettext_from_dict({'en_US': 'en_US'}, 'en', 'da') \
        == 'en_US'
    assert gettext_from_dict({'en': 'en'}, 'en_US', 'da') \
        == 'en'
    assert gettext_from_dict({'en_GB': 'en_GB'}, 'en', 'da') \
        == 'en_GB'
    assert gettext_from_dict({'da': 'da'}, 'de', 'da') \
        == 'da'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for marshmallow GeoJSON Polygon schema."""

import pytest
from marshmallow import ValidationError

from marshmallow_utils.schemas import PolygonSchema


@pytest.mark.parametrize("coordinates", [
    ([[[2.38, 57.322], [23.194, -20.28], [-120.43, 19.15], [2.38, 57.322]]]),
])
def test_polygon(coordinates):
    print(coordinates)
    valid_full = {
        "type": "Polygon",
        "coordinates": coordinates
    }

    loaded = PolygonSchema().load(valid_full)
    # NOTE: Since the schemas return the dict itself, the loaded object
    # is the same than the input and dumped objects (dicts)
    assert valid_full == loaded == PolygonSchema().dump(loaded)


@pytest.mark.parametrize("coordinates", [
    ([2.38, 57.322, -120.43, 19.15]),
    ([[2.38, 57.322], [-120.43, 19.15]]),
    ([[[2.38, 57.322], [23.194, -20.28], [24.194, -19.2]]]),
    ([[[2.38, 57.322], [2.38, 57.322]]]),
    ([[[2.38, 57.322], [23.194, -20.28], [2.38, 57.322]]]),
])
def test_polygon_fail(coordinates):
    invalid_point = {
        "type": "Polygon",
        "coordinates": coordinates
    }
    pytest.raises(ValidationError, PolygonSchema().load, invalid_point)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Link store and field for generating links."""

from marshmallow import fields, missing
from uritemplate import URITemplate


class Links(fields.Field):
    """A links field that knows to look in the context for the schema."""

    # NOTE: forces serialization
    _CHECK_ATTRIBUTE = False

    def _serialize(self, value, attr, obj, *args, **kwargs):
        """Dump the field by using the contextual schema."""
        factory = self.context.get("links_factory")
        namespace = self.context.get("links_namespace")
        schema = (
            factory.get_schema(namespace, self.context) if factory else None
        )
        if schema:
            return schema.dump(obj)
        else:
            return {}


def always(*args, **kwargs):
    """A function that returns True no matter what is passed to it."""
    return True


class Link(fields.Field):
    """A link field that knows how to generate a link from an object."""

    # NOTE: forces serialization
    _CHECK_ATTRIBUTE = False

    def __init__(self, template=None, params=None, permission=None,
                 when=always, **kwargs):
        """Constructor."""
        if isinstance(template, str):
            template = URITemplate(template)
        self.template = template
        self.permission = permission
        self.params = params
        self.when = when
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, *args, **kwargs):
        """Dump the link by using the context."""
        factory = self.context.get("links_factory")
        field_permission_check = self.context.get("field_permission_check")

        if field_permission_check and self.permission:
            if not field_permission_check(self.permission):
                return missing

        if not self.when(obj):
            return missing

        return factory.create_link(self.template, self.params(obj))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""


def test_version():
    """Test version import."""
    from marshmallow_utils import __version__
    assert __version__



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Marshmallow fields."""

from .babel import BabelGettextDictField, FormatDate, FormatDatetime, \
    FormatEDTF, FormatTime
from .contrib import Function, Method
from .edtfdatestring import EDTFDateString
from .generated import GenFunction, GenMethod
from .identifier import IdentifierSet
from .isodate import ISODateString
from .isolanguage import ISOLangString
from .links import Link, Links
from .nestedattr import NestedAttribute
from .sanitizedhtml import ALLOWED_HTML_ATTRS, ALLOWED_HTML_TAGS, SanitizedHTML
from .sanitizedunicode import SanitizedUnicode
from .strippedhtml import StrippedHTML
from .trimmed import TrimmedString
from .tzdatetime import TZDateTime

__all__ = (
    'ALLOWED_HTML_ATTRS',
    'ALLOWED_HTML_TAGS',
    'BabelGettextDictField',
    'EDTFDateString',
    'FormatDate',
    'FormatDatetime',
    'FormatEDTF',
    'FormatTime',
    'Function',
    'GenFunction',
    'GenMethod',
    'IdentifierSet',
    'ISODateString',
    'ISOLangString',
    'Link',
    'Links',
    'Method',
    'NestedAttribute',
    'SanitizedHTML',
    'SanitizedUnicode',
    'StrippedHTML',
    'TrimmedString',
    'TZDateTime',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Field-level permissions."""

from marshmallow import INCLUDE, Schema, fields, post_dump, pre_load, validate


class FieldPermissionError(Exception):
    """Marshmallow field permission error."""


class FieldPermissionsMixin:
    """Mixin for filtering field-level permissions in marshmallow schemas."""

    field_load_permissions = {}
    field_dump_permissions = {}
    default_load_action = None
    default_dump_action = None

    @pre_load
    def _permissions_filter_load(self, data, **kwargs):
        field_permission_check = self.context.get("field_permission_check")
        if field_permission_check:
            for k in self.field_load_permissions:
                if k in data:
                    action = self.field_load_permissions[k] or \
                        self.default_load_action
                    # TODO (Alex): Maybe cache?
                    if action and not field_permission_check(action):
                        raise FieldPermissionError(k)
        return data

    @post_dump
    def _permissions_filter_dump(self, data, **kwargs):
        field_permission_check = self.context.get("field_permission_check")
        # NOTE: we make a copy since we don't want to modify the data when
        # deleting keys
        # TODO (Alex): see if actually needed...
        # data = deepcopy(data)
        if field_permission_check:
            for k in self.field_dump_permissions:
                if k in data:
                    action = self.field_dump_permissions[k] or \
                        self.default_dump_action
                    if action and not field_permission_check(action):
                        del data[k]
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Nested attribute."""

from marshmallow import fields, missing


class AttributeAccessorFieldMixin:
    """Marshmallow field mixin for attribute-based serialization."""

    def get_value(self, obj, attr, accessor=None, default=missing):
        """Return the value for a given key from an object attribute."""
        attribute = getattr(self, "attribute", None)
        check_key = attr if attribute is None else attribute
        return getattr(obj, check_key, default)


class NestedAttribute(AttributeAccessorFieldMixin, fields.Nested):
    """Nested object attribute field."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""GeoJSON Schema."""

from geojson import MultiPoint, Point, Polygon
from marshmallow import Schema, ValidationError
from marshmallow.fields import Constant, Float, List
from marshmallow.validate import Validator
from marshmallow_oneofschema import OneOfSchema


class GeometryValidator(Validator):
    """Validator for GeoJSON geometry objects."""

    def __init__(self, geometry_cls):
        """Initialize the validator.

        :param geometry_cls: The GeoJSON geometry class to validate against.
        """
        self.geometry_cls = geometry_cls

    def __call__(self, value):
        """Validate a geometry object."""
        obj = self.geometry_cls(value)
        if not obj.is_valid:
            errors = obj.errors()
            raise ValidationError({"geojson": {"coordinates": errors}})
        return value


class PointSchema(Schema):
    """GeoJSON Point schema.

    See https://tools.ietf.org/html/rfc7946#section-3.1.2
    """

    coordinates = List(Float, required=True, validate=GeometryValidator(Point))
    type = Constant('Point')


class MultiPointSchema(Schema):
    """GeoJSON MultiPoint schema.

    See https://tools.ietf.org/html/rfc7946#section-3.1.3
    """

    coordinates = List(
        List(Float), required=True, validate=GeometryValidator(MultiPoint))
    type = Constant('MultiPoint')


class PolygonSchema(Schema):
    """GeoJSON Polygon schema.

    See https://tools.ietf.org/html/rfc7946#section-3.1.6
    """

    coordinates = List(
        List(List(Float)), required=True, validate=GeometryValidator(Polygon))
    type = Constant('Polygon')


class GeometryObjectSchema(OneOfSchema):
    """A GeoJSON Geometry Object schema.

    See https://tools.ietf.org/html/rfc7946#section-3.1

    Only Point, MultiPoint and Polygon are supported.
    """

    type_schemas = {
        "Point": PointSchema,
        "MultiPoint": MultiPointSchema,
        "Polygon": PolygonSchema,
    }

    def get_obj_type(self, obj):
        """Finds the object type in the dictionary.

        This function is needed because the schemas return the dict itself.
        Not a geojson object.
        """
        obj_type = obj.get("type")
        if not obj_type or obj_type not in self.type_schemas.keys():
            raise Exception(f"Unknown GeoJSON object type: {obj_type}")

        return obj_type



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for marshmallow Identifiers schema."""

import idutils
import pytest
from marshmallow import ValidationError
from marshmallow.fields import Str

from marshmallow_utils.schemas import IdentifierSchema


def validate_other(identifier):
    """Validates a numeric identifier."""
    return identifier.isnumeric()


dummy_allowed_schemes = {
    "dummy": {"label": "Dummy", "validator": validate_other}
}
#
# Test cases when identifier is not provided:
#


def test_identifier_required_no_value():
    schema = IdentifierSchema(allowed_schemes=dummy_allowed_schemes)
    with pytest.raises(ValidationError) as e:
        schema.load({})

    errors = e.value.normalized_messages()
    assert errors == {'identifier': 'Missing data for required field.',
                      'scheme': 'Missing data for required field.'}


def test_identifier_required_only_scheme():
    schema = IdentifierSchema(allowed_schemes=dummy_allowed_schemes)
    only_scheme = {"scheme": "dummy"}
    with pytest.raises(ValidationError) as e:
        schema.load(only_scheme)

    errors = e.value.normalized_messages()
    assert errors == {'identifier': 'Missing data for required field.'}


def test_identifier_required_empty_value():
    schema = IdentifierSchema(allowed_schemes=dummy_allowed_schemes)
    empty_identifier = {"identifier": "", "scheme": "dummy"}
    with pytest.raises(ValidationError) as e:
        schema.load(empty_identifier)

    errors = e.value.normalized_messages()
    assert errors == {'identifier': 'Missing data for required field.'}


# required, only identifier -> detect schee


def test_identifier_not_required_no_value():
    schema = IdentifierSchema(
        allowed_schemes=dummy_allowed_schemes,
        identifier_required=False
    )
    assert schema.load({}) == {}


def test_identifier_not_required_only_scheme():
    schema = IdentifierSchema(
        allowed_schemes=dummy_allowed_schemes,
        identifier_required=False,
    )
    only_scheme = {"scheme": "dummy"}
    with pytest.raises(ValidationError) as e:
        schema.load(only_scheme)

    errors = e.value.normalized_messages()
    assert errors == {'identifier': 'Missing data for required field.'}


#
# Test cases when identifier provided:
#

# This schema accepts two attributes: scheme and identifier. The use
# cases are provided in the following table, `-` means that the value
# does not apply.

# 1- The scheme is given -> is allowed (idutils) -> validate -> pass
# 2- The scheme is given -> is NOT allowed -> fail
# 3- The scheme is given -> allowed is empty -> fail
# 4- The scheme is given -> is allowed (custom)  -> validate -> pass
# 5- The scheme is NOT given -> is detected -> is allowed -> validate -> pass
# 6- The scheme is NOT given -> is detected -> is NOT allowed -> fail
# 7- The scheme is NOT given -> is detected -> allowed is empty -> fail
# 8- The scheme is NOT given -> is detected (several) -> is allowed (one of them) -> pass  # noqa
# 9- The scheme is NOT given -> is detected (several) -> is allowed (not the first) -> pass  # noqa
# 10- The scheme is NOT given -> is NOT detected -> fail

# | given  | detected |
# | scheme |  scheme  |  alllowed schemes   | given identifier | result |
# |:------:|:--------:|:-------------------:|:----------------:|:------:|
# |   DOI  |     -    |        [DOI]        | 10.12345/foo.bar |  pass  |
# |   DOI  |     -    |       [Other]       | 10.12345/foo.bar |  fail  |
# |   DOI  |     -    |         []          | 10.12345/foo.bar |  fail  |
# |  Other |     -    | [(Other, val_func)] |     123456       |  pass  |
# |    -   |   [DOI]  |        [DOI]        | 10.12345/foo.bar |  pass  |
# |    -   |   [DOI]  |       [Other]       | 10.12345/foo.bar |  fail  |
# |    -   |   [DOI]  |         []          | 10.12345/foo.bar |  fail  |
# |    -   |  [isni,  |       [isni]        |     0317-8471    |  pass  |  (scheme = isni)  # noqa
# |        |  orcid]  |                     |                  |        |
# |    -   |  [isni,  |       [orcid]       |     0317-8471    |  pass  |  (scheme = orcid)  # noqa
# |        |  orcid]  |                     |                  |        |
# |    -   |     -    |       [isni]        |    00:11:22:33   |  fail  |


def test_given_and_allowed_scheme_valid_value():  # 1
    allowed_schemes = {
        "doi": {"label": "DOI", "validator": idutils.is_doi}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_doi = {
        "scheme": "doi",
        "identifier": "10.12345/foo.bar"
    }
    data = schema.load(valid_doi)
    assert data == valid_doi


def test_given_and_allowed_scheme_invalid_value():  # 1
    allowed_schemes = {
        "doi": {"label": "DOI", "validator": idutils.is_doi}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    invalid_doi = {
        "scheme": "doi",
        "identifier": "12345"
    }
    with pytest.raises(ValidationError) as e:
        schema.load(invalid_doi)

    errors = e.value.normalized_messages()
    assert errors == {'identifier': 'Invalid DOI identifier.'}


def test_given_and_not_allowed_scheme_valid_value():  # 2
    allowed_schemes = {
        "other": {"label": "Other", "validator": validate_other}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_doi = {
        "scheme": "doi",
        "identifier": "10.12345/foo.bar"
    }
    with pytest.raises(ValidationError) as e:
        schema.load(valid_doi)

    errors = e.value.normalized_messages()
    assert errors == {'scheme': 'Invalid scheme.'}


def test_given_custom_and_allowed_scheme_valid_value():  # 4
    allowed_schemes = {
        "other": {"label": "Other", "validator": validate_other}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_other = {
        "scheme": "other",
        "identifier": "12345"
    }

    data = schema.load(valid_other)
    assert data == valid_other


def test_given_custom_and_allowed_scheme_invalid_value():  # 4
    allowed_schemes = {
        "other": {"label": "Other", "validator": validate_other}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    invalid_other = {
        "scheme": "other",
        "identifier": "12345abc"
    }

    with pytest.raises(ValidationError) as e:
        schema.load(invalid_other)

    errors = e.value.normalized_messages()
    assert errors == {
        'identifier': 'Invalid Other identifier.'
    }


def test_detected_and_allowed_scheme_valid_value():  # 5
    allowed_schemes = {
        "doi": {"label": "DOI", "validator": idutils.is_doi}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_doi = {
        "identifier": "10.12345/foo.bar"
    }
    data = schema.load(valid_doi)

    valid_doi["scheme"] = "doi"
    assert data == valid_doi


def test_detected_and_not_allowed_scheme_valid_value():  # 6
    allowed_schemes = {
        "other": {"label": "Other", "validator": validate_other}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_doi = {
        "identifier": "10.12345/foo.bar"
    }

    with pytest.raises(ValidationError) as e:
        schema.load(valid_doi)

    errors = e.value.normalized_messages()
    assert errors == {
        'scheme': 'Missing data for required field.'
    }


def test_detected_and_allowed_scheme_respect_detection_order():  # 8
    allowed_schemes = {
        "orcid": {"label": "ORCID", "validator": idutils.is_orcid},
        "isni": {"label": "ISNI", "validator": idutils.is_isni}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_orcid = {
        "identifier": "0000-0001-6759-6273"
    }
    data = schema.load(valid_orcid)

    valid_orcid["scheme"] = "orcid"
    assert data == valid_orcid


def test_detected_and_allowed_scheme_second_detected():  # 8
    allowed_schemes = {
        "isni": {"label": "ISNI", "validator": idutils.is_isni}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    valid_isni = {
        "identifier": "0000-0001-6759-6273"
    }
    data = schema.load(valid_isni)

    valid_isni["scheme"] = "isni"
    assert data == valid_isni


def test_not_given_not_detected_scheme_for_identifier():  # 10
    allowed_schemes = {
        "isni": {"label": "ISNI", "validator": idutils.is_isni}
    }
    schema = IdentifierSchema(allowed_schemes=allowed_schemes)
    invalid_no_scheme = {
        "identifier": "00:11:22:33"
    }

    with pytest.raises(ValidationError) as e:
        schema.load(invalid_no_scheme)

    errors = e.value.normalized_messages()
    assert errors == {
        'scheme': 'Missing data for required field.'
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

r"""Extras and utilities for Marshmallow.

Currently this library contains a couple of extra fields that helps with
sanitizing data as shown in the following example:

>>> from marshmallow_utils import fields
>>> from marshmallow import Schema
>>> class MySchema(Schema):
...     trim = fields.TrimmedString()
...     html = fields.SanitizedHTML()
...     text = fields.SanitizedUnicode()
...     isodate = fields.ISODateString()
...
>>> data = MySchema().load({
...    'trim': '    whitespace   ',
...    'html': '<script>evil()</script>',
...    'text': 'PDF copy/paste\u200b\u000b\u001b\u0018 ',
...    'isodate': '1999-10-27',
... })
>>> data['trim']
'whitespace'
>>> data['html']
'evil()'
>>> data['text']
'PDF copy/paste'
>>> data['isodate']
'1999-10-27'

Fields:

- :py:class:`~fields.SanitizedUnicode`: Integrates the
  `ftfy <https://pypi.org/project/ftfy/>`_ for fixing broken unicode text.
- :py:class:`~fields.SanitizedHTML`: Integrates the
  `bleach <https://pypi.org/project/bleach/>`_ for HTML sanitization.
- :py:class:`~fields.ISODateString`: Integrates the
  `arrow <https://pypi.org/project/arrow/>`_ for date parsing.
"""

from .version import __version__

__all__ = ('__version__', )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Extras and utilities for Marshmallow"""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'check-manifest>=0.42',
    'coverage>=5.2.1',
    'pytest-cov>=2.10.1',
    'pytest-isort>=1.2.0',
    'pytest-pycodestyle>=2.2.0',
    'pytest-pydocstyle>=2.2.0',
    'pytest>=6.0',
]

setup_requires = [
    'Babel>=2.8,<3',
]

extras_require = {
    'docs': [
        'Sphinx>=4.2.0',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

install_requires = [
    'arrow>=0.16.0',
    'babel-edtf>=1.0.0',
    'bleach>=3.0.0',
    'edtf>=4.0.1,<5.0.0',
    'ftfy>=4.4.3',
    'geojson>=2.5.0',
    'idutils>=1.1.8',
    'marshmallow>=3.5.0,<4.0.0',
    'marshmallow-oneofschema>=2.1.0',
    'pycountry>=18.12.8',
    'uritemplate>=3.0.1',
    'werkzeug>=1.0.0',
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('marshmallow_utils', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='marshmallow-utils',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='marshmallow utils extras',
    license='MIT',
    author='CERN',
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/marshmallow-utils',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
        entry_points={
        'invenio_i18n.translations': [
            'marshmallow_utils = marshmallow_utils',
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 5 - Production/Stable',
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""HTML sanitized string field."""

from marshmallow import fields

from ..html import strip_html


class StrippedHTML(fields.String):
    """String field which strips HTML entities.

    The value is stripped using the bleach library. Any already escaped value
    is being unescaped before return.
    """

    def _serialize(self, value, attr, data, **kwargs):
        """Serialize string by stripping HTML entities."""
        value = super()._serialize(
            value, attr, data, **kwargs)
        return strip_html(value)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Version information for Marshmallow-Utils.

This file is imported by ``marshmallow_utils.__init__``,
and parsed by ``setup.py``.
"""

__version__ = '0.5.5'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""HTML sanitized string field."""

from marshmallow import fields

# For backward compatibility we import ALLOWED_* variables.
from ..html import ALLOWED_HTML_ATTRS, ALLOWED_HTML_TAGS, sanitize_html


class SanitizedHTML(fields.String):
    """String field which sanitizes HTML using the bleach library.

    The default list of allowed tags and attributes is defined by
    ``ALLOWED_HTML_TAGS`` and ``ALLOWED_HTML_ATTRS``.

    You can override the defaults like this:

    .. code-block:: python

        class MySchema(Schema):
            html = fields.SanitizedHTML(tags=['a'], attrs={'a': ['href']})

    :param tags: List of allowed tags.
    :param attrs: Dictionary of allowed attributes per tag.
    """

    def __init__(self, tags=None, attrs=None, *args, **kwargs):
        """Initialize field."""
        super().__init__(*args, **kwargs)
        self.tags = tags
        self.attrs = attrs

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize string by sanitizing HTML."""
        value = super()._deserialize(
            value, attr, data, **kwargs)
        return sanitize_html(value, tags=self.tags, attrs=self.attrs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identifier field."""

from marshmallow import ValidationError
from marshmallow.fields import List


class IdentifierSet(List):
    """Identifier list with deduplication.

    It assumes the items of the list contain a *scheme* property.
    """

    default_error_messages = {
        "multiple_values": "Only one identifier per scheme is allowed.",
    }

    def _validate(self, value):
        """Validates the list of identifiers."""
        schemes = [identifier["scheme"] for identifier in value]
        if not len(value) == len(set(schemes)):
            raise self.make_error(key="multiple_values")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Date string field."""

import arrow
from arrow.parser import ParserError
from marshmallow import fields, missing


class ISODateString(fields.Date):
    """ISO8601-formatted date string.

    ISODateString serializes to a date string and if it can't, the field is
    ignored (missing).

    NOTE: It serializes None to None.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        """Serialize an ISO8601-formatted date."""
        # WHY: arrow.get(None) returns a date but we don't want it to
        if value is None:
            return missing

        try:
            return super()._serialize(
                arrow.get(value).date(), attr, obj, **kwargs)
        except ParserError:
            return missing

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize an ISO8601-formatted date."""
        return super()._deserialize(value, attr, data, **kwargs).isoformat()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""ISO language string field."""

import pycountry
from marshmallow import ValidationError

from .sanitizedunicode import SanitizedUnicode


def _(x):
    return x


def validate_iso639_3(value):
    """Validate that language is ISO 639-3 value."""
    if not pycountry.languages.get(alpha_3=value):
        raise ValidationError(
            _('Language must be a lower-cased 3-letter ISO 639-3 string.'),
            field_name=['language']
        )


class ISOLangString(SanitizedUnicode):
    """ISO language string field."""

    def __init__(self, validate=validate_iso639_3, *args, **kwargs):
        """ISO language string field initialization."""
        super(ISOLangString, self).__init__(
            validate=validate, *args, **kwargs
        )
